@@ -15,10 +15,8 @@ RETRY_INTERVAL = 2  # deciseconds
 STORAGE_KEY = "supabase.auth.token"
 
 API_VERSION_HEADER_NAME = "X-Supabase-Api-Version"
-API_VERSIONS = {
-    "2024-01-01": {
-        "timestamp": datetime.timestamp(datetime.strptime("2024-01-01", "%Y-%m-%d")),
-        "name": "2024-01-01",
-    },
-}
+API_VERSIONS_2024_01_01_TIMESTAMP = datetime.timestamp(
+    datetime.strptime("2024-01-01", "%Y-%m-%d")
+)
+API_VERSIONS_2024_01_01_NAME = "2024-01-01"
 BASE64URL_REGEX = r"^([a-z0-9_-]{4})*($|[a-z0-9_-]{3}$|[a-z0-9_-]{2}$)$"
